//! A 256-bit non-cryptographic hash function.
//!
//! This is intended to be used as a fast, high-quality checksum for
//! non-adversarial data identification.  It is not intended to stand
//! up to attacks of any kind.
//!
//! This uses the MIX function and permutation patterns from Skein v1.3,
//! but is otherwise unrelated.  For example, it is not tweakable and
//! uses far fewer rounds per data chunk.
//!
//! This implementation assumes support for 64-bit unsigned integers.
//!
//! This implementation should work on platforms of any endianness,
//! but has only been tested on little endian platforms.  Running the
//! unit tests on a big-endian platform can verify.

const BLOCK_SIZE: usize = 256 / 8; // Block size of the hash, in bytes.
const UPDATE_MIX_ROUNDS: usize = 2;
const FINISH_MIX_ROUNDS: usize = 6;

/// Convenience function to generate a hash for a block of data.
pub fn hash(data: &[u8]) -> [u8; BLOCK_SIZE] {
    let mut h = LedHash256::new();
    h.update(data);
    h.finish()
}

/// A hash builder.  Consumes bytes and generates a 256-bit hash.
#[derive(Debug, Copy, Clone)]
#[repr(C)]
#[repr(align(32))]
pub struct LedHash256 {
    state: [u64; 4],       // Hash state.
    buf: [u8; BLOCK_SIZE], // Accumulates message data for processing.
    buf_length: usize,     // The number of message bytes currently stored in buf[].
    message_length: u64,   // Accumulates the total message length, in bytes.
}

impl LedHash256 {
    pub fn new() -> LedHash256 {
        LedHash256 {
            state: [
                0x302f7bfd333d0b0d,
                0xb98cf5312d92a4d5,
                0xb1c885da4c257a6e,
                0xedb85f57b82c7e10,
            ],
            buf: [0; BLOCK_SIZE],
            buf_length: 0,
            message_length: 0,
        }
    }

    /// Update the hash with new data.
    pub fn update(&mut self, data: &[u8]) {
        self.message_length += data.len() as u64;

        let mut data = data;
        while !data.is_empty() {
            if self.buf_length == BLOCK_SIZE {
                // Process the filled buffer.
                add_buffer_to_state(&mut self.state, &self.buf);
                mix_state(&mut self.state, UPDATE_MIX_ROUNDS);
                self.buf_length = 0;
            } else if self.buf_length == 0 && data.len() >= BLOCK_SIZE {
                // Process data directly, skipping the buffer.
                add_buffer_to_state(&mut self.state, data);
                mix_state(&mut self.state, UPDATE_MIX_ROUNDS);
                data = &data[BLOCK_SIZE..];
            } else {
                // Fill the buffer.
                let n = (BLOCK_SIZE - self.buf_length).min(data.len());
                (&mut self.buf[self.buf_length..(self.buf_length + n)]).copy_from_slice(&data[..n]);
                data = &data[n..];
                self.buf_length += n;
            }
        }
    }

    /// Finishes the hash calculations and returns the digest.
    pub fn finish(mut self) -> [u8; BLOCK_SIZE] {
        // Hash the remaining bytes if there are any.
        if self.buf_length > 0 {
            (&mut self.buf[self.buf_length..]).fill(0);
            add_buffer_to_state(&mut self.state, &self.buf);
            mix_state(&mut self.state, UPDATE_MIX_ROUNDS);
        }

        // Incorporate the message length (in bits) and do the
        // final mixing.
        self.state[0] ^= self.message_length * 8;
        mix_state(&mut self.state, FINISH_MIX_ROUNDS);

        // Get the digest as a byte array and return it.
        let mut digest = [0u8; BLOCK_SIZE];
        digest[0..8].copy_from_slice(&self.state[0].to_le_bytes());
        digest[8..16].copy_from_slice(&self.state[1].to_le_bytes());
        digest[16..24].copy_from_slice(&self.state[2].to_le_bytes());
        digest[24..32].copy_from_slice(&self.state[3].to_le_bytes());
        return digest;
    }
}

/// Adds the contents of a buffer to the hash state.
///
/// The buffer must be at least 32 bytes long.  Only the first 32 bytes
/// are added.
#[inline(always)]
fn add_buffer_to_state(state: &mut [u64; 4], buffer: &[u8]) {
    use std::convert::TryInto;

    // Convert the buffer to native endian u64's and xor into the
    // hash state.
    assert!(buffer.len() >= BLOCK_SIZE);
    state[0] ^= u64::from_le_bytes((&buffer[0..8]).try_into().unwrap());
    state[1] ^= u64::from_le_bytes((&buffer[8..16]).try_into().unwrap());
    state[2] ^= u64::from_le_bytes((&buffer[16..24]).try_into().unwrap());
    state[3] ^= u64::from_le_bytes((&buffer[24..32]).try_into().unwrap());
}

/// The main mix function.  Mixes the passed hash state.
///
/// Inspired by Skein 1.3, and using its MIX function.
#[inline(always)]
fn mix_state(state: &mut [u64; 4], rounds: usize) {
    const ROTATIONS: &[[u32; 4]] = &[
        [40, 50, 27, 21],
        [40, 50, 27, 21],
        [40, 50, 27, 21],
        [40, 50, 27, 21],
    ];
    for round in 0..rounds {
        let rot = ROTATIONS[round % ROTATIONS.len()];

        // MIX function.
        state[0] = state[0].wrapping_add(state[2]);
        state[2] = state[2].rotate_left(rot[0]) ^ state[0];
        state[1] = state[1].wrapping_add(state[3]);
        state[3] = state[3].rotate_left(rot[1]) ^ state[1];

        // We flip the indices we use below, as if we did
        // a [0 1 2 3] -> [0 1 3 2] permutation.

        // MIX function.
        state[0] = state[0].wrapping_add(state[3]);
        state[3] = state[3].rotate_left(rot[2]) ^ state[0];
        state[1] = state[1].wrapping_add(state[2]);
        state[2] = state[2].rotate_left(rot[3]) ^ state[1];
    }
}

#[cfg(test)]
mod test {
    use super::*;

    fn digest_to_string(digest: [u8; 32]) -> String {
        fn low_bits_to_char(n: u8) -> char {
            match n {
                0 => '0',
                1 => '1',
                2 => '2',
                3 => '3',
                4 => '4',
                5 => '5',
                6 => '6',
                7 => '7',
                8 => '8',
                9 => '9',
                10 => 'a',
                11 => 'b',
                12 => 'c',
                13 => 'd',
                14 => 'e',
                15 => 'f',
                _ => unreachable!(),
            }
        }

        let mut s = String::new();
        for byte in &digest {
            s.push(low_bits_to_char(byte >> 4u8));
            s.push(low_bits_to_char(byte & 0b00001111));
        }
        s
    }

    #[test]
    fn hash_empty() {
        let correct_digest = "3c26b703d4d7316c7bc6fc8e72893433cb10e1044fb5100ac89d2a67c85ff1a1";
        assert_eq!(digest_to_string(hash(&[])), correct_digest);
    }

    #[test]
    fn hash_zero() {
        let correct_digest = "65670a6fd5d2e919f3d5ec856dd18d7c66397683e47f059ac19f2533b1607708";
        assert_eq!(digest_to_string(hash(&[0u8])), correct_digest);
    }

    #[test]
    fn hash_one() {
        let correct_digest = "6666ad08b0d0164f1972f36be7541c04b6e22947710d52b2733dcd3007908b82";
        assert_eq!(digest_to_string(hash(&[1u8])), correct_digest);
    }

    #[test]
    fn hash_string_01() {
        let s = "abc";
        let correct_digest = "891862c9ad6003a588194f7b59c6760d48083faa5eda5ade03b5eb0551c538e6";
        assert_eq!(digest_to_string(hash(s.as_bytes())), correct_digest);
    }

    #[test]
    fn hash_string_02() {
        let s = "The quick brown fox jumps over the lazy dog.";
        let correct_digest = "347cce28ceb133b8ab30cca93e52460c40d30156ddf9122b3e13a6239c1e3f17";
        assert_eq!(digest_to_string(hash(s.as_bytes())), correct_digest);
    }

    #[test]
    fn hash_string_03() {
        let s = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789";
        let correct_digest = "99b7eeae84e40ef9371b398af09e4cad00e800417b74b45f5af6042b49d03e1e";
        assert_eq!(digest_to_string(hash(s.as_bytes())), correct_digest);
    }

    #[test]
    fn hash_string_04() {
        let s = "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.";
        let correct_digest = "8d84fe8c5bc2a41ba0f0f57013bef7f038acf1d2d6a77ebc4d1b0fa14c10629c";
        assert_eq!(digest_to_string(hash(s.as_bytes())), correct_digest);
    }

    #[test]
    fn hash_multi_part_processing() {
        let test_string1 =
            "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do eiusmod tempor";
        let test_string2 = " incididunt ut l";
        let test_string3 = "abore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat ";
        let test_string4 = "cup";
        let test_string5 =
            "idatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.";
        let correct_digest = "8d84fe8c5bc2a41ba0f0f57013bef7f038acf1d2d6a77ebc4d1b0fa14c10629c";

        let mut hasher = LedHash256::new();
        hasher.update(test_string1.as_bytes());
        hasher.update(test_string2.as_bytes());
        hasher.update(test_string3.as_bytes());
        hasher.update(test_string4.as_bytes());
        hasher.update(test_string5.as_bytes());
        let digest = hasher.finish();

        assert_eq!(digest_to_string(digest), correct_digest);
    }

    #[test]
    fn hash_length() {
        // We're testing here to make sure the length of the data properly
        // affects the hash.  Internally in the hash, the last block of data
        // is padded with zeros, so here we're forcing that last block to be
        // all zeros, and only changing the length of input.
        let len_0 = &[];
        let len_1 = &[0u8];
        let len_2 = &[0u8, 0];

        assert_eq!(
            digest_to_string(hash(len_0)),
            "3c26b703d4d7316c7bc6fc8e72893433cb10e1044fb5100ac89d2a67c85ff1a1",
        );
        assert_eq!(
            digest_to_string(hash(len_1)),
            "65670a6fd5d2e919f3d5ec856dd18d7c66397683e47f059ac19f2533b1607708",
        );
        assert_eq!(
            digest_to_string(hash(len_2)),
            "95bba56b21bdaf4a6c8c3a231c4966c0992845757a73fa6bbd48389cf8b7b452",
        );
    }
}
